#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
from unittest import TestCase
from parameterized import parameterized
from numpy import (nan)

from pandas import (
    Series,
    DataFrame,
    date_range,
    MultiIndex,
    Timedelta,
    Timestamp,
    concat,
)
from pandas.util.testing import (assert_frame_equal,
                                 assert_series_equal)

from .. utils import (get_clean_factor_and_forward_returns,
                      compute_forward_returns,
                      quantize_factor)


class UtilsTestCase(TestCase):
    dr = date_range(start='2015-1-1', end='2015-1-2')
    dr.name = 'date'
    tickers = ['A', 'B', 'C', 'D']

    factor = DataFrame(index=dr,
                       columns=tickers,
                       data=[[1, 2, 3, 4],
                             [4, 3, 2, 1]]).stack()
    factor.index = factor.index.set_names(['date', 'asset'])
    factor.name = 'factor'
    factor_data = DataFrame()
    factor_data['factor'] = factor
    factor_data['group'] = Series(index=factor.index,
                                  data=[1, 1, 2, 2, 1, 1, 2, 2],
                                  dtype="category")

    biased_factor = DataFrame(index=dr,
                              columns=tickers.extend(['E', 'F', 'G', 'H']),
                              data=[[-1, 3, -2, 4, -5, 7, -6, 8],
                                    [-4, 2, -3, 1, -8, 6, -7, 5]]).stack()
    biased_factor.index = biased_factor.index.set_names(['date', 'asset'])
    biased_factor.name = 'factor'
    biased_factor_data = DataFrame()
    biased_factor_data['factor'] = biased_factor
    biased_factor_data['group'] = Series(index=biased_factor.index,
                                         data=[1, 1, 2, 2, 1, 1, 2, 2,
                                               1, 1, 2, 2, 1, 1, 2, 2],
                                         dtype="category")

    def test_compute_forward_returns(self):
        dr = date_range(start='2015-1-1', end='2015-1-3')
        prices = DataFrame(index=dr, columns=['A', 'B'],
                           data=[[1, 1], [1, 2], [2, 1]])
        factor = prices.stack()

        fp = compute_forward_returns(factor, prices, periods=[1, 2])

        ix = MultiIndex.from_product([dr, ['A', 'B']],
                                     names=['date', 'asset'])
        expected = DataFrame(index=ix, columns=['1D', '2D'])
        expected['1D'] = [0., 1., 1., -0.5, nan, nan]
        expected['2D'] = [1., 0., nan, nan, nan, nan]

        assert_frame_equal(fp, expected)

    def test_compute_forward_returns_index_out_of_bound(self):
        dr = date_range(start='2014-12-29', end='2015-1-3')
        prices = DataFrame(index=dr, columns=['A', 'B'],
                           data=[[nan, nan], [nan, nan], [nan, nan],
                                 [1, 1], [1, 2], [2, 1]])

        dr = date_range(start='2015-1-1', end='2015-1-3')
        factor = DataFrame(index=dr, columns=['A', 'B'],
                           data=[[1, 1], [1, 2], [2, 1]])
        factor = factor.stack()

        fp = compute_forward_returns(factor, prices, periods=[1, 2])

        ix = MultiIndex.from_product([dr, ['A', 'B']],
                                     names=['date', 'asset'])
        expected = DataFrame(index=ix, columns=['1D', '2D'])
        expected['1D'] = [0., 1., 1., -0.5, nan, nan]
        expected['2D'] = [1., 0., nan, nan, nan, nan]

        assert_frame_equal(fp, expected)

    def test_compute_forward_returns_non_cum(self):
        dr = date_range(start='2015-1-1', end='2015-1-3')
        prices = DataFrame(index=dr, columns=['A', 'B'],
                           data=[[1, 1], [1, 2], [2, 1]])
        factor = prices.stack()

        fp = compute_forward_returns(factor, prices, periods=[1, 2],
                                     cumulative_returns=False)

        ix = MultiIndex.from_product([dr, ['A', 'B']],
                                     names=['date', 'asset'])
        expected = DataFrame(index=ix, columns=['1D', '2D'])
        expected['1D'] = [0., 1., 1., -0.5, nan, nan]
        expected['2D'] = [1., -0.5, nan, nan, nan, nan]

        assert_frame_equal(fp, expected)

    @parameterized.expand([(factor_data, 4, None, False, False,
                            [1, 2, 3, 4, 4, 3, 2, 1]),
                           (factor_data, 2, None, False, False,
                            [1, 1, 2, 2, 2, 2, 1, 1]),
                           (factor_data, 2, None, True, False,
                            [1, 2, 1, 2, 2, 1, 2, 1]),
                           (biased_factor_data, 4, None, False, True,
                            [2, 3, 2, 3, 1, 4, 1, 4, 2, 3, 2, 3, 1, 4, 1, 4]),
                           (biased_factor_data, 2, None, False, True,
                            [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2]),
                           (biased_factor_data, 2, None, True, True,
                            [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2]),
                           (biased_factor_data, None, 4, False, True,
                            [2, 3, 2, 3, 1, 4, 1, 4, 2, 3, 2, 3, 1, 4, 1, 4]),
                           (biased_factor_data, None, 2, False, True,
                            [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2]),
                           (biased_factor_data, None, 2, True, True,
                            [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2]),
                           (factor_data, [0, .25, .5, .75, 1.], None, False,
                            False, [1, 2, 3, 4, 4, 3, 2, 1]),
                           (factor_data, [0, .5, .75, 1.], None, False, False,
                            [1, 1, 2, 3, 3, 2, 1, 1]),
                           (factor_data, [0, .25, .5, 1.], None, False, False,
                            [1, 2, 3, 3, 3, 3, 2, 1]),
                           (factor_data, [0, .5, 1.], None, False, False,
                            [1, 1, 2, 2, 2, 2, 1, 1]),
                           (factor_data, [.25, .5, .75], None, False, False,
                            [nan, 1, 2, nan, nan, 2, 1, nan]),
                           (factor_data, [0, .5, 1.], None, True, False,
                            [1, 2, 1, 2, 2, 1, 2, 1]),
                           (factor_data, [.5, 1.], None, True, False,
                            [nan, 1, nan, 1, 1, nan, 1, nan]),
                           (factor_data, [0, 1.], None, True, False,
                            [1, 1, 1, 1, 1, 1, 1, 1]),
                           (factor_data, None, 4, False, False,
                            [1, 2, 3, 4, 4, 3, 2, 1]),
                           (factor_data, None, 2, False, False,
                            [1, 1, 2, 2, 2, 2, 1, 1]),
                           (factor_data, None, 3, False, False,
                            [1, 1, 2, 3, 3, 2, 1, 1]),
                           (factor_data, None, 8, False, False,
                            [1, 3, 6, 8, 8, 6, 3, 1]),
                           (factor_data, None, [0, 1, 2, 3, 5], False, False,
                            [1, 2, 3, 4, 4, 3, 2, 1]),
                           (factor_data, None, [1, 2, 3], False, False,
                            [nan, 1, 2, nan, nan, 2, 1, nan]),
                           (factor_data, None, [0, 2, 5], False, False,
                            [1, 1, 2, 2, 2, 2, 1, 1]),
                           (factor_data, None, [0.5, 2.5, 4.5], False, False,
                            [1, 1, 2, 2, 2, 2, 1, 1]),
                           (factor_data, None, [0.5, 2.5], True, False,
                            [1, 1, nan, nan, nan, nan, 1, 1]),
                           (factor_data, None, 2, True, False,
                            [1, 2, 1, 2, 2, 1, 2, 1])])
    def test_quantize_factor(self, factor, quantiles, bins, by_group,
                             zero_aware, expected_vals):
        quantized_factor = quantize_factor(factor,
                                           quantiles=quantiles,
                                           bins=bins,
                                           by_group=by_group,
                                           zero_aware=zero_aware)
        expected = Series(index=factor.index,
                          data=expected_vals,
                          name='factor_quantile').dropna()
        assert_series_equal(quantized_factor, expected)

    def test_get_clean_factor_and_forward_returns_1(self):
        """
        Test get_clean_factor_and_forward_returns with a daily factor
        """
        tickers = ['A', 'B', 'C', 'D', 'E', 'F']

        factor_groups = {'A': 1, 'B': 2, 'C': 1, 'D': 2, 'E': 1, 'F': 2}

        price_data = [[1.10**i, 0.50**i, 3.00**i, 0.90**i, 0.50**i, 1.00**i]
                      for i in range(1, 7)]  # 6 days = 3 + 3 fwd returns

        factor_data = [[3, 4, 2, 1, nan, nan],
                       [3, nan, nan, 1, 4, 2],
                       [3, 4, 2, 1, nan, nan]]  # 3 days

        start = '2015-1-11'
        factor_end = '2015-1-13'
        price_end = '2015-1-16'  # 3D fwd returns

        price_index = date_range(start=start, end=price_end)
        price_index.name = 'date'
        prices = DataFrame(index=price_index, columns=tickers, data=price_data)

        factor_index = date_range(start=start, end=factor_end)
        factor_index.name = 'date'
        factor = DataFrame(index=factor_index, columns=tickers,
                           data=factor_data).stack()

        factor_data = get_clean_factor_and_forward_returns(
            factor, prices,
            groupby=factor_groups,
            quantiles=4,
            periods=(1, 2, 3))

        expected_idx = factor.index.rename(['date', 'asset'])
        expected_cols = ['1D', '2D', '3D',
                         'factor', 'group', 'factor_quantile']
        expected_data = [[0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.5, -0.75, -0.875, 4.0, 2, 4],
                         [2.0,  8.00, 26.000, 2.0, 1, 2],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1],
                         [0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1],
                         [-0.5, -0.75, -0.875, 4.0, 1, 4],
                         [0.0,  0.00,  0.000, 2.0, 2, 2],
                         [0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.5, -0.75, -0.875, 4.0, 2, 4],
                         [2.0,  8.00, 26.000, 2.0, 1, 2],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1]]
        expected = DataFrame(index=expected_idx,
                             columns=expected_cols, data=expected_data)
        expected['group'] = expected['group'].astype('category')

        assert_frame_equal(factor_data, expected)

    def test_get_clean_factor_and_forward_returns_2(self):
        """
        Test get_clean_factor_and_forward_returns with a daily factor
        on a business day calendar
        """
        tickers = ['A', 'B', 'C', 'D', 'E', 'F']

        factor_groups = {'A': 1, 'B': 2, 'C': 1, 'D': 2, 'E': 1, 'F': 2}

        price_data = [[1.10**i, 0.50**i, 3.00**i, 0.90**i, 0.50**i, 1.00**i]
                      for i in range(1, 7)]  # 6 days = 3 + 3 fwd returns

        factor_data = [[3, 4, 2, 1, nan, nan],
                       [3, nan, nan, 1, 4, 2],
                       [3, 4, 2, 1, nan, nan]]  # 3 days

        start = '2017-1-12'
        factor_end = '2017-1-16'
        price_end = '2017-1-19'  # 3D fwd returns

        price_index = date_range(start=start, end=price_end, freq='B')
        price_index.name = 'date'
        prices = DataFrame(index=price_index, columns=tickers, data=price_data)

        factor_index = date_range(start=start, end=factor_end, freq='B')
        factor_index.name = 'date'
        factor = DataFrame(index=factor_index, columns=tickers,
                           data=factor_data).stack()

        factor_data = get_clean_factor_and_forward_returns(
            factor, prices,
            groupby=factor_groups,
            quantiles=4,
            periods=(1, 2, 3))

        expected_idx = factor.index.rename(['date', 'asset'])
        expected_cols = ['1D', '2D', '3D',
                         'factor', 'group', 'factor_quantile']
        expected_data = [[0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.5, -0.75, -0.875, 4.0, 2, 4],
                         [2.0,  8.00, 26.000, 2.0, 1, 2],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1],
                         [0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1],
                         [-0.5, -0.75, -0.875, 4.0, 1, 4],
                         [0.0,  0.00,  0.000, 2.0, 2, 2],
                         [0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.5, -0.75, -0.875, 4.0, 2, 4],
                         [2.0,  8.00, 26.000, 2.0, 1, 2],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1]]
        expected = DataFrame(index=expected_idx,
                             columns=expected_cols, data=expected_data)
        expected['group'] = expected['group'].astype('category')

        assert_frame_equal(factor_data, expected)

    def test_get_clean_factor_and_forward_returns_3(self):
        """
        Test get_clean_factor_and_forward_returns with and intraday factor
        """
        tickers = ['A', 'B', 'C', 'D', 'E', 'F']

        factor_groups = {'A': 1, 'B': 2, 'C': 1, 'D': 2, 'E': 1, 'F': 2}

        price_data = [[1.10**i, 0.50**i, 3.00**i, 0.90**i, 0.50**i, 1.00**i]
                      for i in range(1, 5)]  # 4 days = 3 + 1 fwd returns

        factor_data = [[3, 4, 2, 1, nan, nan],
                       [3, nan, nan, 1, 4, 2],
                       [3, 4, 2, 1, nan, nan]]  # 3 days

        start = '2017-1-12'
        factor_end = '2017-1-16'
        price_end = '2017-1-17'  # 1D fwd returns

        price_index = date_range(start=start, end=price_end, freq='B')
        price_index.name = 'date'
        today_open = DataFrame(index=price_index + Timedelta('9h30m'),
                               columns=tickers, data=price_data)
        today_open_1h = DataFrame(index=price_index + Timedelta('10h30m'),
                                  columns=tickers, data=price_data)
        today_open_1h += today_open_1h * 0.001
        today_open_3h = DataFrame(index=price_index + Timedelta('12h30m'),
                                  columns=tickers, data=price_data)
        today_open_3h -= today_open_3h * 0.002
        prices = concat([today_open, today_open_1h, today_open_3h]) \
            .sort_index()

        factor_index = date_range(start=start, end=factor_end, freq='B')
        factor_index.name = 'date'
        factor = DataFrame(index=factor_index + Timedelta('9h30m'),
                           columns=tickers, data=factor_data).stack()

        factor_data = get_clean_factor_and_forward_returns(
            factor, prices,
            groupby=factor_groups,
            quantiles=4,
            periods=(1, 2, 3))

        expected_idx = factor.index.rename(['date', 'asset'])
        expected_cols = ['1h', '3h', '1D',
                         'factor', 'group', 'factor_quantile']
        expected_data = [[0.001, -0.002, 0.1, 3.0, 1, 3],
                         [0.001, -0.002, -0.5, 4.0, 2, 4],
                         [0.001, -0.002, 2.0, 2.0, 1, 2],
                         [0.001, -0.002, -0.1, 1.0, 2, 1],
                         [0.001, -0.002, 0.1, 3.0, 1, 3],
                         [0.001, -0.002, -0.1, 1.0, 2, 1],
                         [0.001, -0.002, -0.5, 4.0, 1, 4],
                         [0.001, -0.002, 0.0, 2.0, 2, 2],
                         [0.001, -0.002, 0.1, 3.0, 1, 3],
                         [0.001, -0.002, -0.5, 4.0, 2, 4],
                         [0.001, -0.002, 2.0, 2.0, 1, 2],
                         [0.001, -0.002, -0.1, 1.0, 2, 1]]
        expected = DataFrame(index=expected_idx,
                             columns=expected_cols, data=expected_data)
        expected['group'] = expected['group'].astype('category')

        assert_frame_equal(factor_data, expected)

    def test_get_clean_factor_and_forward_returns_4(self):
        """
        Test get_clean_factor_and_forward_returns on an event
        """
        tickers = ['A', 'B', 'C', 'D', 'E', 'F']

        factor_groups = {'A': 1, 'B': 2, 'C': 1, 'D': 2, 'E': 1, 'F': 2}

        price_data = [[1.10**i, 0.50**i, 3.00**i, 0.90**i, 0.50**i, 1.00**i]
                      for i in range(1, 9)]

        factor_data = [[1, nan, nan, nan, nan, 6],
                       [4, nan, nan, 7, nan, nan],
                       [nan, nan, nan, nan, nan, nan],
                       [nan, 3, nan, 2, nan, nan],
                       [nan, nan, 1, nan, 3, nan]]

        price_index = date_range(start='2017-1-12', end='2017-1-23', freq='B')
        price_index.name = 'date'
        prices = DataFrame(index=price_index, columns=tickers, data=price_data)

        factor_index = date_range(start='2017-1-12', end='2017-1-18', freq='B')
        factor_index.name = 'date'
        factor = DataFrame(index=factor_index, columns=tickers,
                           data=factor_data).stack()

        factor_data = get_clean_factor_and_forward_returns(
            factor, prices,
            groupby=factor_groups,
            quantiles=4,
            periods=(1, 2, 3))

        expected_idx = factor.index.rename(['date', 'asset'])
        expected_cols = ['1D', '2D', '3D',
                         'factor', 'group', 'factor_quantile']
        expected_data = [[0.1,  0.21,  0.331, 1.0, 1, 1],
                         [0.0,   0.00,  0.000, 6.0, 2, 4],
                         [0.1,  0.21,  0.331, 4.0, 1, 1],
                         [-0.1, -0.19, -0.271, 7.0, 2, 4],
                         [-0.5, -0.75, -0.875, 3.0, 2, 4],
                         [-0.1, -0.19, -0.271, 2.0, 2, 1],
                         [2.0,  8.00, 26.000, 1.0, 1, 1],
                         [-0.5, -0.75, -0.875, 3.0, 1, 4]]
        expected = DataFrame(index=expected_idx,
                             columns=expected_cols, data=expected_data)
        expected['group'] = expected['group'].astype('category')

        assert_frame_equal(factor_data, expected)

    def test_get_clean_factor_and_forward_returns_5(self):
        """
        Test get_clean_factor_and_forward_returns with and intraday factor
        and holidays
        """
        tickers = ['A', 'B', 'C', 'D', 'E', 'F']

        factor_groups = {'A': 1, 'B': 2, 'C': 1, 'D': 2, 'E': 1, 'F': 2}

        price_data = [[1.10**i, 0.50**i, 3.00**i, 0.90**i, 0.50**i, 1.00**i]
                      for i in range(1, 20)]  # 19 days = 18 + 1 fwd returns

        factor_data = [[3, 4, 2, 1, nan, nan],
                       [3, nan, nan, 1, 4, 2],
                       [3, 4, 2, 1, nan, nan]] * 6  # 18 days

        start = '2017-1-12'
        factor_end = '2017-2-10'
        price_end = '2017-2-13'  # 1D (business day) fwd returns
        holidays = ['2017-1-13', '2017-1-18', '2017-1-30', '2017-2-7']
        holidays = [Timestamp(d) for d in holidays]

        price_index = date_range(start=start, end=price_end, freq='B')
        price_index.name = 'date'
        price_index = price_index.drop(holidays)

        today_open = DataFrame(index=price_index + Timedelta('9h30m'),
                               columns=tickers, data=price_data)
        today_open_1h = DataFrame(index=price_index + Timedelta('10h30m'),
                                  columns=tickers, data=price_data)
        today_open_1h += today_open_1h * 0.001
        today_open_3h = DataFrame(index=price_index + Timedelta('12h30m'),
                                  columns=tickers, data=price_data)
        today_open_3h -= today_open_3h * 0.002
        prices = concat([today_open, today_open_1h, today_open_3h]) \
            .sort_index()

        factor_index = date_range(start=start, end=factor_end, freq='B')
        factor_index.name = 'date'
        factor_index = factor_index.drop(holidays)
        factor = DataFrame(index=factor_index + Timedelta('9h30m'),
                           columns=tickers, data=factor_data).stack()

        factor_data = get_clean_factor_and_forward_returns(
            factor, prices,
            groupby=factor_groups,
            quantiles=4,
            periods=(1, 2, 3))

        expected_idx = factor.index.rename(['date', 'asset'])
        expected_cols = ['1h', '3h', '1D',
                         'factor', 'group', 'factor_quantile']
        expected_data = [[0.001, -0.002, 0.1, 3.0, 1, 3],
                         [0.001, -0.002, -0.5, 4.0, 2, 4],
                         [0.001, -0.002, 2.0, 2.0, 1, 2],
                         [0.001, -0.002, -0.1, 1.0, 2, 1],
                         [0.001, -0.002, 0.1, 3.0, 1, 3],
                         [0.001, -0.002, -0.1, 1.0, 2, 1],
                         [0.001, -0.002, -0.5, 4.0, 1, 4],
                         [0.001, -0.002, 0.0, 2.0, 2, 2],
                         [0.001, -0.002, 0.1, 3.0, 1, 3],
                         [0.001, -0.002, -0.5, 4.0, 2, 4],
                         [0.001, -0.002, 2.0, 2.0, 1, 2],
                         [0.001, -0.002, -0.1, 1.0, 2, 1]] * 6  # 18  days
        expected = DataFrame(index=expected_idx,
                             columns=expected_cols, data=expected_data)
        expected['group'] = expected['group'].astype('category')

        assert_frame_equal(factor_data, expected)

        inferred_holidays = factor_data.index.levels[0].freq.holidays
        assert sorted(holidays) == sorted(inferred_holidays)

    def test_get_clean_factor_and_forward_returns_6(self):
        """
        Test get_clean_factor_and_forward_returns with a daily factor
        on a business day calendar and holidays
        """
        tickers = ['A', 'B', 'C', 'D', 'E', 'F']

        factor_groups = {'A': 1, 'B': 2, 'C': 1, 'D': 2, 'E': 1, 'F': 2}

        price_data = [[1.10**i, 0.50**i, 3.00**i, 0.90**i, 0.50**i, 1.00**i]
                      for i in range(1, 22)]  # 21 days = 18 + 3 fwd returns

        factor_data = [[3, 4, 2, 1, nan, nan],
                       [3, nan, nan, 1, 4, 2],
                       [3, 4, 2, 1, nan, nan]] * 6  # 18 days

        start = '2017-1-12'
        factor_end = '2017-2-10'
        price_end = '2017-2-15'  # 3D (business day) fwd returns
        holidays = ['2017-1-13', '2017-1-18', '2017-1-30', '2017-2-7']
        holidays = [Timestamp(d) for d in holidays]

        price_index = date_range(start=start, end=price_end, freq='B')
        price_index.name = 'date'
        price_index = price_index.drop(holidays)
        prices = DataFrame(index=price_index, columns=tickers, data=price_data)

        factor_index = date_range(start=start, end=factor_end, freq='B')
        factor_index.name = 'date'
        factor_index = factor_index.drop(holidays)
        factor = DataFrame(index=factor_index, columns=tickers,
                           data=factor_data).stack()

        factor_data = get_clean_factor_and_forward_returns(
            factor, prices,
            groupby=factor_groups,
            quantiles=4,
            periods=(1, 2, 3))

        expected_idx = factor.index.rename(['date', 'asset'])
        expected_cols = ['1D', '2D', '3D',
                         'factor', 'group', 'factor_quantile']
        expected_data = [[0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.5, -0.75, -0.875, 4.0, 2, 4],
                         [2.0,  8.00, 26.000, 2.0, 1, 2],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1],
                         [0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1],
                         [-0.5, -0.75, -0.875, 4.0, 1, 4],
                         [0.0,  0.00,  0.000, 2.0, 2, 2],
                         [0.1,  0.21,  0.331, 3.0, 1, 3],
                         [-0.5, -0.75, -0.875, 4.0, 2, 4],
                         [2.0,  8.00, 26.000, 2.0, 1, 2],
                         [-0.1, -0.19, -0.271, 1.0, 2, 1]] * 6  # 18  days
        expected = DataFrame(index=expected_idx,
                             columns=expected_cols, data=expected_data)
        expected['group'] = expected['group'].astype('category')

        assert_frame_equal(factor_data, expected)

        inferred_holidays = factor_data.index.levels[0].freq.holidays
        assert sorted(holidays) == sorted(inferred_holidays)



from . import performance
from . import plotting
from . import api as tears
from . import utils


from ._version import get_versions


__version__ = get_versions()['version']
del get_versions

__all__ = ['performance', 'plotting','utils'].extend(tears.__all__)



import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

import empyrical as ep
from .alphalens.tears import sd_get_Analysis_results
from .alphalens.utils import get_clean_factor_and_forward_returns
import warnings

# from ..dataget import pro


warnings.filterwarnings("ignore")

def rolling_window(a, window):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

def decay(df, window, mode='linear', exp_f=1, NaToZero=False):
    """
    加权衰减函数，用于减少换手，时序函数，常用即放于此处
    线性加权:x[date]*window + x[date-1]*(window-1) + x[date-2]*(window-2) + x[date-3]*(window-3)...
    指数加权0<exp_f<1:x[date] + x[date-1]*exp_f^1 + x[date-2]*exp_f^2 + x[date-3]**exp_f^3...
    """

    if mode == 'linear':
        weight = np.array(range(1, window+1))
    elif mode == 'exp':
        weight = np.array(range(window-1, -1, -1))
        weight = exp_f**weight
    elif mode == 'exp2':
        weight = 2 ** (-((window - np.linspace(1, window, window) + 1) / (window/2)))
    else:
        raise ValueError('请确认加权模式')

    if NaToZero:
        nr = rolling_window(~np.isnan(df.values.T), window).copy()
        sum_w = df.copy()
        sum_w[:] = np.NaN
        sum_w.iloc[window-1:] = np.dot(nr, weight).T
        sum_w = sum_w.replace(0, np.nan)
        df = df.fillna(0)
    else:
        sum_w = weight.sum()

    r = rolling_window(df.values.T, window).copy()
    res = df.copy()
    res[:] = np.NaN
    res.iloc[window-1:] = np.dot(r, weight).T

    res = res/sum_w
    return res

def filter_stock_byDate(close, filterdays_pre, filterdays_end):
    """
    按上市和退市时间标记可用时间为1，不可用为nan
    Parameters
    ----------
    close : 带有date和issue的时间序列.
    filterdays_pre : 上市filterdays_pre之后置标记
    filterdays_end : 退市filterdays_pre之前置标记

    Returns
    -------
    date,issue,filterFlag

    """
    close = close.astype('float64')
    close[close < 0.01] = np.nan
    filterFlag_pre = close.shift(filterdays_pre)
    filterFlag_pre.iloc[:filterdays_pre] = close.iloc[0]
    filterFlag_end = close.shift(-filterdays_end)
    filterFlag_end.iloc[-filterdays_end:] = close.iloc[-1]

    filterFlag = close*np.nan
    filterFlag[close.notna()] = 1
    filterFlag[filterFlag_pre.isna()] = np.nan
    filterFlag[filterFlag_end.isna()] = np.nan

    return filterFlag



START_DATE = '19900101'
END_DATE = '20900101'



def wash_daily_price_factor(daily,factor,start_date=None,end_date=None, is_limit=True):
    """根据具有date, issue, close, adj, is_limit_buy, is_limit_sell的DataFrame清洗价格和因子"""
    if start_date is None:
        start_date = START_DATE
    if end_date is None:
        end_date = END_DATE
    
    daily = daily[(daily['date'] >= start_date) & (daily['date'] <= end_date)]
    daily['close_adj'] = daily['close']*daily['adj'] 
    close = daily.pivot(index='date', columns='issue', values='close_adj').sort_index()
    close.columns.name = 'asset'
    
    # 上市日期限制
    filterFlag = filter_stock_byDate(close, 90, 90)
    factor = factor * filterFlag    
    
    # 涨跌停限制
    if is_limit:
        is_limit_buy = daily.pivot(index='date', columns='issue', values='is_limit_buy')
        is_limit_sell = daily.pivot(index='date', columns='issue', values='is_limit_sell')

        factor[is_limit_buy == 1] = np.nan  # 涨停不选
        factor[is_limit_sell == 1] = np.nan   # 跌停不选
    
    factor = factor.stack().rename('factor').dropna()
    factor.index.names = ['date','asset']
    return close,factor 

def analysis(factor,prices,is_limit_buy,is_limit_sell,benchmark, quantiles,plot=False, periods=(1,3,5,10),mode='longonly'):
    """
    因子分析
    :param start_datetime开始时间
    :param end_datetime结束时间
    :param factor_data因子数据
    :param price_data量价数据
    """

    if mode=='longonly':
        print("*" * 50 + '\n' + '------输出做多收益模式指标')
    if mode=='benchmark':
        print("*" * 50 + '\n' + '------输出对标超额收益模式指标')
    if mode=='longshort':
        print("*" * 50 + '\n' + '------输出多空收益模式指标')

    factorRtndata = get_clean_factor_and_forward_returns(factor=factor, 
                                                         prices=prices,
                                                         quantiles=quantiles,
                                                         periods=periods, 
                                                         max_loss=0.1, 
                                                         benchmark=benchmark)
    results = sd_get_Analysis_results(factorRtndata, ana_mode=mode, plot=plot)
    print('Alphalens Complete')

    def sd_get_Nav_results(factor, prices, is_limit_buy, is_limit_sell,num=100, bench=None):
        factor = factor.unstack().sort_index()
        close = prices.replace([0, np.inf, -np.inf], np.nan)
        close = close.loc[close.index >= max(factor.index.min(), close.index.min())]
        close = close.loc[close.index <= min(factor.index.max(), close.index.max())]
        # ret = price_data['ret'].reindex_like(close).replace([np.nan, np.inf, -np.inf], 0)
        ret = close.pct_change(fill_method=None).replace([np.nan, np.inf, -np.inf], 0)
        is_limit_buy = is_limit_buy.reindex_like(close).fillna(1).astype(int)
        is_limit_sell = is_limit_sell.reindex_like(close).fillna(1).astype(int)
        factor = factor.reindex_like(close).astype('float64')
        if bench is None:
            benchmark = pd.Series(data=0, index=close.index)
        else:
            benchmark = bench.reindex(close.index).pct_change(fill_method=None).fillna(0)
        # %
        final_Report = dict()
        final_Report['weights'] = close.fillna(0)*0
        final_Report['dailyResults'] = pd.DataFrame(index=close.index)
        final_Report['dailyResults'].loc[:, 'dailyreturn'] = 0
        final_Report['dailyResults'].loc[:, 'dailyturnover'] = 0

        weight = final_Report['weights'].iloc[0, :].fillna(0)
        for d in range(len(close.index)):
            pre_date = close.index[d-1] if d > 0 else 0
            date = close.index[d]

            con_buylimit = is_limit_buy.loc[date] == 1
            con_selllimit = is_limit_sell.loc[date] == 1
            con_noexists = close.loc[date] < 0.01
            con_trade = (~ con_selllimit) & (~ con_buylimit) & (~ con_noexists)

            score = factor.loc[date]
            # score[~con_trade] = np.nan
            score = score[con_trade]
            score = score.dropna()
            # score = score.rename('score')
            pre_weight = final_Report['weights'].loc[pre_date].fillna(0).rename('weight') if d > 0 else final_Report['weights'].loc[date].fillna(0).rename('weight')
            pre_weight = pre_weight[pre_weight > 0.0001]

            n = int(min(num, np.floor(len(score)/1+0.5)))
            score = score.nlargest(n)
            weight = final_Report['weights'].loc[pre_date] if d > 0 else final_Report['weights'].loc[date]
            pre_weight = weight
            if len(score) > 0:
                weight = pre_weight*0
                weight[score.index] = 1/len(score)

            weight[con_selllimit] = np.maximum(weight[con_selllimit], pre_weight[con_selllimit])
            weight[con_buylimit] = np.minimum(weight[con_buylimit], pre_weight[con_buylimit])
            weight[con_selllimit & con_buylimit] = pre_weight[con_selllimit & con_buylimit]
            weight[con_noexists] = 0
            if weight[con_trade].sum() > 0.001:
                weight[con_trade] = weight[con_trade]*max(0, (1-weight[con_buylimit | con_buylimit].sum()))
            else:
                weight[con_trade] = 0

            final_Report['dailyResults'].loc[date, 'dailyreturn'] = np.sum(
                pre_weight*ret.loc[date])-benchmark.loc[date]
            final_Report['dailyResults'].loc[date, 'dailyturnover'] = sum(
                abs(weight-pre_weight))
            final_Report['weights'].loc[date] = weight
        final_Report['dailyResults']['dailyreturnAFTfees'] = final_Report['dailyResults']['dailyreturn'] - \
            final_Report['dailyResults']['dailyturnover']/2*0.0013
        if bench is None:
            final_Report['dailyResults']['return'] = (1+final_Report['dailyResults']['dailyreturn']).cumprod()
            final_Report['dailyResults']['returnAFTfees'] = (1+final_Report['dailyResults']['dailyreturnAFTfees']).cumprod()
        else:
            final_Report['dailyResults']['return'] = final_Report['dailyResults']['dailyreturn'].cumsum()
            final_Report['dailyResults']['returnAFTfees'] = final_Report['dailyResults']['dailyreturnAFTfees'].cumsum()

        Summary_table = pd.DataFrame()
        trading_days = len(close)
        Summary_table.loc['Annualized Return(%)', 'value'] = final_Report['dailyResults']['return'] .iloc[-1]/trading_days*243*100
        Summary_table.loc['Annualized ReturnAE(%)', 'value'] = final_Report['dailyResults']['returnAFTfees'] .iloc[-1]/trading_days*243*100
        Summary_table.loc['max_drawdown(%)', 'value'] = ep.max_drawdown(final_Report['dailyResults']['dailyreturnAFTfees'])*100
        Summary_table.loc['SharpeRatio', 'value'] = ep.sharpe_ratio(final_Report['dailyResults']['dailyreturnAFTfees'])
        Summary_table.loc['WinP20', 'value'] = (final_Report['dailyResults']['dailyreturnAFTfees'].rolling(20).sum() >= 0).sum()/(trading_days-20)
        Summary_table.loc['WinP60', 'value'] = (final_Report['dailyResults']['dailyreturnAFTfees'].rolling(60).sum() >= 0).sum()/(trading_days-60)
        Summary_table.loc['Mean turnover', 'value'] = final_Report['dailyResults']['dailyturnover'].mean()/2
        Summary_table.loc['Mean position', 'value'] = final_Report['weights'].sum(axis=1).mean()
        final_Report['Summary_table'] = Summary_table
        plotdata = pd.merge(final_Report['dailyResults']['return'],
                            final_Report['dailyResults']['returnAFTfees'], left_index=True, right_index=True)

        if bench is None:

            mean_quant_rateret_byYear = (final_Report['dailyResults']['dailyreturn'].add(1).groupby(final_Report['dailyResults']['dailyreturn'].index.get_level_values('date').to_period('A')).prod()**(
                243/final_Report['dailyResults']['dailyreturn'].add(1).groupby(final_Report['dailyResults']['dailyreturn'].index.get_level_values('date').to_period('A')).count())).add(-1)
            mean_quant_rateret_byYear_ae = (final_Report['dailyResults']['dailyreturnAFTfees'].add(1).groupby(final_Report['dailyResults']['dailyreturnAFTfees'].index.get_level_values('date').to_period('A')).prod()**(
                243/final_Report['dailyResults']['dailyreturnAFTfees'].add(1).groupby(final_Report['dailyResults']['returnAFTfees'].index.get_level_values('date').to_period('A')).count())).add(-1)
        else:
            mean_quant_rateret_byYear = final_Report['dailyResults']['dailyreturn'].groupby(final_Report['dailyResults']['dailyreturn'].index.get_level_values('date').to_period('A')).mean()*243
            mean_quant_rateret_byYear_ae = final_Report['dailyResults']['dailyreturnAFTfees'].groupby(final_Report['dailyResults']['dailyreturnAFTfees'].index.get_level_values('date').to_period('A')).mean()*243

        returns_table_byYear = pd.DataFrame()
        returns_table_byYear["Return(%)"] = (mean_quant_rateret_byYear * 100).round(2)
        returns_table_byYear["ReturnAE(%)"] = (mean_quant_rateret_byYear_ae * 100).round(2)
        returns_table_byYear["SR"] = \
            final_Report['dailyResults']['dailyreturnAFTfees'].groupby(final_Report['dailyResults']['dailyreturnAFTfees'].index.get_level_values('date').to_period('A'), group_keys=False).apply(ep.sharpe_ratio).round(2)
        returns_table_byYear["mdd(%)"] = \
            final_Report['dailyResults']['dailyreturnAFTfees'].groupby(final_Report['dailyResults']['dailyreturnAFTfees'].index.get_level_values('date').to_period('A'), group_keys=False).apply(ep.max_drawdown).round(4) * 100

        final_Report['returns_table_byYear'] = returns_table_byYear
        return final_Report, plotdata

    results_nav, plotdata_nav = sd_get_Nav_results(factor, prices, is_limit_buy,is_limit_sell,100, benchmark)
    results['results_nav'] = results_nav
    return results



class SingleFactorAnalysis:
    """
    prices: DataFrame, index = date, columns = asset
    factor: Series, Multiindex, level_0 = date, level_1 = asset
    is_limit_buy: prices like
    is_limit_sell: prices like
    benchmark: str|Series
    
    results:
        results = {}
        results['keyResults'] = keyResults_table:DataFrame
        results['ret_Data'] = ret_Data:dict
            'returns_table': 分组收益率统计量
            'ret_quant_daily': 分组日收益序列
            'ret_spread_quant_daily': 不同频率多空收益
            'ret_quant_daily_ae': 费后分组日收益
            'ret_spread_quant_daily_ae': 费后多空收益
            'cum_ret': 累计分组收益
            'ret_wide': 1D分组日收益
            'cum_ret_ae': 累计费后分组收益
            'ret_wide_ae': 1D费后分组日收益
            'returns_table_byYear':收益率结果按年统计
            'quantile_return':分组年化收益
            'quantile_return_ae':费后分组年化收益
        results['turnover_Data'] = turnover_Data:dict
            'turnover_table': 统计回测区间内总的分组换手率
            'quantile_turnover': 具体到某天的分组换手率
        results['IC_Data'] = IC_Data:DataFrame
    """
    
    def __init__(self,start_date,end_date,factor,prices,is_limit_buy,is_limit_sell,benchmark=None,period=(1,3,5,10),quantiles=None):
        self._start_date = start_date
        self._end_date = end_date
        self._factor = factor.loc[start_date:end_date]
        self._prices = prices
        self._is_limit_buy = is_limit_buy
        self._is_limit_sell = is_limit_sell
        self._benchmark = benchmark
        self._period = period
        if quantiles is None:
            quantiles = int(factor.groupby(level='date').count().mean()/(100)+0.5)
        self._quantiles=quantiles
        self.result = {}    
    def longonly_mode(self):
        if 'longonly' in self.result.keys():
            return self.result['longonly']
        self.result['longonly'] = analysis(self._factor,
                                           self._prices,
                                           self._is_limit_buy,
                                           self._is_limit_sell,
                                           quantiles=self._quantiles,
                                           benchmark=None,
                                           periods=self._period)
        return self.result['longonly']
    
    def benchmark_mode(self):
        if 'bench' in self.result.keys():
            return self.result['bench']
        if self._benchmark is None:
            raise ValueError("Need to provide benchmark price data")
        # if isinstance(self._benchmark, str):
        #     benchmark = pro.index_daily(self._benchmark)['close']
        # else:
        #     benchmark = self._benchmark
            
        self.result['bench'] = analysis(self._factor,
                                        self._prices,
                                        self._is_limit_buy,
                                        self._is_limit_sell,
                                        quantiles=self._quantiles,
                                        periods=self._period,
                                        benchmark=self._benchmark,
                                        mode='benchmark')
        return self.result['bench']
    
    def longshort_mode(self):
        if 'longshort' in self.result.keys():
            return self.result['longshort']
        self.result['longshort'] = analysis(self._factor,
                                           self._prices,
                                           self._is_limit_buy,
                                           self._is_limit_sell,
                                           quantiles=self._quantiles,
                                           benchmark=None,
                                           periods=self._period)
        return self.result['longshort']    

    
    


import numpy as np
import pandas as pd
from tqdm import tqdm
import copy as copy
import math
from collections import Counter
from sklearn.linear_model import LinearRegression

# 注：所有data1均为dataframe格式，不是series，不是series

def stable_place(xx):
    """
    划分成若干等长区间，计算各值所处的区间中值，返回众数
    :param xx:
    :return:
    """
    try:
        ranges=0.005
        top=xx.max()
        buttom=xx.min()
        denominator=xx.mean()
        n=math.ceil((top-buttom)/denominator/ranges)+1
        ran=np.linspace(buttom,top,n)
        x=copy.deepcopy(xx.values)
        for num in range(len(x)):
            y=x[num]
            for i in range(len(ran)-1):
                down=ran[i]
                up=ran[i+1]
                if (y>=down)&(y<up):
                    x[num]=(up+down)/2
                    break
        count=Counter(x)
        result=count.most_common(1)[0][0]
    except:
        result=0
    return result

####################时间序列####################
def Ts_Single_Rate(data1,n):
    """
    计算data1中每一列在时序上的变化百分比，基数为n前数

    注：会有大数出现
    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    result = data1.pct_change(periods = n).fillna(0)
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result

def Ts_Single_AccRate(data1,n):
    """
    计算data1中每一列在时序上的变化百分比，
    基数为n前数，计算至今每一期的变化率，并求和，Acc: accumulate
    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    result = data1.rolling(n,min_periods=2).sum()/data1.shift(n) - n
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})

    return result

def Ts_Shape_Compare(data1,data2,n):
    """
    对比data1与data2中每一列的值，回溯n期，统计n期内data2大于data1的次数

    Parameters
    ----------
    data1: Dataframe
    data2: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    data1_values=data1.values
    data2_values=data2.values
    result = pd.DataFrame(data1_values < data2_values, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result

def Ts_Increase_Times_Discontinuous(data1,n):
    """
    回溯期n，计算是增长的次数

    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    result = (data1.diff() > 0).rolling(n,min_periods=2).sum()
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})

    return result

def Ts_Decrease_Times_Discontinuous(data1,n):
    """
    回溯期n，计算是增长的次数

    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    result = (data1.diff() < 0).rolling(n,min_periods=2).sum()
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})

    return result

def Ts_Unchanged_Times_Discontinuous(data1,n):
    """
    回溯期n，计算是增长的次数

    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    result = (data1.diff() == 0).rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_Jump(data1, n):
    """
    回溯期n，期间出现不同值的次数

    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """

    result = data1.rolling(n,min_periods=2).apply(lambda x: len(x.unique()))
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Ts_Single_StablePoint(data1,n):
    result = data1.rolling(n,min_periods=2).apply(stable_place)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

## todo 计算中的参数设定存疑
def Ts_Single_Sharpee(data1,n):
    """
    回溯期n，夏普除以标准差

    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    rolling_mean = data1.rolling(n,min_periods=2).mean()
    rolling_std = data1.rolling(n,min_periods=2).std(ddof=0)** (3/2)
    result = rolling_mean / rolling_std * np.sqrt(250/n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_Sharp(data1,n):
    """
    回溯期n，夏普

    Parameters
    ----------
    data1: Dataframe
    n:

    Return
    ----------
    result: Dataframe
    """
    rolling_mean = data1.rolling(n,min_periods=2).mean()
    rolling_std = data1.rolling(n,min_periods=2).std(ddof = 0)
    result = rolling_mean / rolling_std * np.sqrt(250/n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_UpTrendRatio(data1,n):
    if n==1:
        result=copy.deepcopy(data1)
    else:
        result=np.full(data1.shape,np.nan)
        data1_values=data1.values
        start=n-1
        end=data1.shape[0]
        parameters=n-1
        for i in tqdm(range(start,end)):
            temp=data1_values[i-n+1:i+1,:]
            total_trend=((temp[-1]-temp[1])>0).astype(int)
            # 向量化运算
            count = np.nansum(np.diff(temp,axis = 0) >0 ,axis = 0)
            #
            rough_trend=count/parameters
            result[i,:]=total_trend*rough_trend

        where_are_inf = np.isinf(result)

        result[where_are_inf] = np.nan
        result[np.isnan(data1)] = np.nan
        result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_DownTrendRatio(data1,n):
    if n==1:
        result=copy.deepcopy(data1)
    else:
        result=np.full(data1.shape,np.nan)
        data1_values=data1.values
        start=n-1
        end=data1.shape[0]
        parameters=n-1
        for i in tqdm(range(start,end)):
            temp=data1_values[i-n+1:i+1,:]
            total_trend=((temp[-1]-temp[1])<0).astype(int)
            #
            count = np.nansum(np.diff(temp, axis=0) < 0, axis=0)
            #
            rough_trend=count/parameters
            result[i,:]=total_trend*rough_trend

        where_are_inf = np.isinf(result)

        result[where_are_inf] = np.nan
        result[np.isnan(data1)] = np.nan
        result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

# 计算均值等方法略有差异，效率未能提升，多列使用矩阵运算较快。
# def Ts_Single_AutoCorrelation(data1,n,lag=1):
#     result = data1.rolling(n,min_periods=2).apply(lambda x:x.autocorr(lag = lag))
#     result = result.where(~np.isinf(result), np.nan)
#   
#     result = result.where(~np.isinf(result), np.nan)
#     return result
def Ts_Single_AutoCorrelation(data1,n,lag=1):
    #lag=1
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=lag+n-1
    end=data1.shape[0]
    result = []
    for i in range(start):
        result.append(np.full(data1.shape[1],np.nan))
    for i in tqdm(range(start,end)):
        temp1 = data1_values[i-(n-1):i+1,:]
        temp2 = data1_values[i-(n-1)-lag:i+1-lag,:]
        
        num = np.nanmean(((temp1 - np.nanmean(temp1,axis = 0))*(temp2-np.nanmean(temp2,axis = 0))),axis = 0)
        den = (np.nanstd(temp1,axis = 0 )*np.nanstd(temp2,axis = 0))
        den = np.where(den == 0, np.nan, den)
        result.append(num/den)
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result

def Ts_Single_AutoCorrelation1(data1,n):
    lag=1
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=lag+n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1 = data1_values[i-(n-1):i+1,:]
        temp2 = data1_values[i-(n-1)-lag:i+1-lag,:]
        result[i,:] = np.nanmean(((temp1 - np.nanmean(temp1,axis = 0))*(temp2-np.nanmean(temp2,axis = 0))),axis = 0)/((np.nanstd(temp1,axis = 0 )*np.nanstd(temp2,axis = 0)))
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result

def Ts_Single_ExpSmooth1(data1,n):
    alpha=0.7
    if n==1:
        result=copy.deepcopy(data1)
    else:
        wt = np.append((1 - alpha) ** (n - 1), (1 - alpha) ** np.arange(0, n - 1)[::-1] * alpha)
        result=np.zeros(data1.shape)
        data1_values=data1.values
        start=n
        end=data1.shape[0]
        for i in tqdm(range(start,end)):
            temp=data1_values[i-n+1:i+1,:]
            result[i,:]= np.dot(wt,temp)

        where_are_inf = np.isinf(result)

        result[where_are_inf] = np.nan
        result[np.isnan(data1)] = 0
        result = pd.DataFrame(result, index=data1.index, columns=data1.columns)
    return result

def Ts_Single_PositiveN(data1,n):
    positive_mark = data1 > 0
    result = positive_mark.rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_RSI(data1,n):
    data_pos = data1.where(data1 > 0, 0)
    data_abs = np.abs(data1)
    result = data_pos.rolling(n,min_periods=2).sum() / data_abs.rolling(n,min_periods=2).sum() * 100
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Min(data1,n):
    result = data1.rolling(n,min_periods=2).min()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Max(data1,n):
    result = data1.rolling(n,min_periods=2).max()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Mean(data1,n):
    result = data1.rolling(n,min_periods=2).mean()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Sum(data1,n):
    result = data1.rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Multiply(data1,n):
    result = data1 * data1.shift(n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Divide(data1,n):
    result = data1 / data1.shift(n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Delay(data1,n):
    result = data1.shift(n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Delta(data1,n):
    result = data1.diff(periods = n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Plus(data1,n):
    result = data1 - data1.shift(n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Stddev(data1,n):
    result = data1.rolling(n,min_periods=2).std(ddof = 0)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

# todo skew和kurt略有不同，大差不差
def Ts_Single_Skewness(data1,n):
    result = data1.rolling(n,min_periods=2).skew()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Ts_Single_Kurtosis(data1,n):
    result = data1.rolling(n,min_periods=2).kurt()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

## todo TS 尚未改进

def Ts_Increase_Times_Continuous(data1,n):
    """
    回溯期n，最大连续增长的次数

    未能优化
    """
    data1_values=data1.values
    result=np.zeros(data1.shape)
    start=n
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        df=[]
        data_temp=data1_values[i-n:i+1,]
        count_times=np.zeros(data1.shape[1])
        for j in range(1,n+1):
            symbol=(data_temp[j]>data_temp[j-1])
            count_times+=symbol.astype(int)
            df.append(copy.deepcopy(count_times))
            count_times[~symbol]=0
        result[i,:]+=np.max(df,axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

# def Ts_Increase_Times_Continuous(data1,n):
#     """
#     回溯期n，最大连续增长的次数
#
#     Parameters
#     ----------
#     data1: Dataframe
#     n:
#
#     Return
#     ----------
#     result: Dataframe
#     """
#
#     def longest_continuous_one(x):
#         max_count = 0
#         count = 0
#         for i in x:
#             if i == 1:
#                 count += 1
#             else:
#                 max_count = max(max_count, count)
#                 count = 0
#         max_count = max(max_count, count)
#         return max_count
#     result = (data1.diff()>0).rolling(n,min_periods=2)
#     result = result.apply(longest_continuous_one)
#     result = result.where(~np.isinf(result), np.nan)
#   
#     result = result.where(~np.isinf(result), np.nan)
#     return result

def Ts_Decrease_Times_Continuous(data1,n):
    data1_values=data1.values
    result=np.zeros(data1.shape)
    start=n
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        df=[]
        data_temp=data1_values[i-n:i+1,]
        count_times=np.zeros(data1.shape[1])
        for j in range(1,n+1):
            symbol=(data_temp[j]<data_temp[j-1])
            count_times+=symbol.astype(int)
            df.append(copy.deepcopy(count_times))
            count_times[~symbol]=0
        result[i,:]+=np.max(df,axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Unchanged_Times_Continuous(data1,n):
    data1_values=data1.values
    result=np.zeros(data1.shape)
    start=n
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        df=[]
        data_temp=data1_values[i-n:i+1,]
        count_times=np.zeros(data1.shape[1])
        for j in range(1,n+1):
            symbol=(data_temp[j]==data_temp[j-1])
            count_times+=symbol.astype(int)
            df.append(copy.deepcopy(count_times))
            count_times[~symbol]=0
        result[i,:]+=np.max(df,axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_ExpSmooth2(data1,n):
    alpha=0.7
    beta=0.7
    if n==1:
        result=copy.deepcopy(data1)
    else:
        result=np.zeros(data1.shape)
        data1_values=data1.values
        start=n
        end=data1.shape[0]
        for i in tqdm(range(start,end)):
            temp=data1_values[i-n+1:i+1,:]
            temp1=temp[0,:]
            semp1=np.zeros(data1_values.shape[1])
            for j in range(1,n):
                temp0=copy.deepcopy(temp1)
                temp1=(1-alpha)*(temp1+semp1)+(alpha)*(temp[j])
                semp1=beta*(temp1-temp0)+(1-beta)*semp1
            result[i,:]=temp1

        where_are_inf = np.isinf(result)

        result[where_are_inf] = np.nan
        result[np.isnan(data1)] = np.nan
        result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_Pressure(data1,n):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1=copy.deepcopy(data1_values[i-n+1:i+1,:])
        matrix=(temp1>=temp1[-1,:]).astype(int)
        mul=temp1*matrix
        mul[mul==0]=np.nan
        result[i,:]=np.nanmedian(mul,axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_Support(data1,n):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1=copy.deepcopy(data1_values[i-n+1:i+1,:])
        matrix=(temp1<=temp1[-1,:]).astype(int)
        mul=temp1*matrix
        mul[mul==0]=np.nan
        result[i,:]=np.nanmedian(mul,axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_UpBoll(data1,n,k):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1=copy.deepcopy(data1_values[i-n+1:i+1,:])
        result[i,:]=data1_values[i,:]+np.nanstd(temp1,axis=0)*k
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_DownBoll(data1,n,k):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1=copy.deepcopy(data1_values[i-n+1:i+1,:])
        result[i,:]=data1_values[i,:]-np.nanstd(temp1,axis=0)*k
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_AroonUp(data1,n):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1=copy.deepcopy(data1_values[i-n+1:i+1,:])
        result[i,:]=(n-np.argmax(temp1,axis=0))/n*100
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_AroonDown(data1,n):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1=copy.deepcopy(data1_values[i-n+1:i+1,:])
        result[i,:]=(n-np.argmin(temp1,axis=0))/n*100
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result


def Ts_Single_Prod(data1,n):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        result[i,:]=np.nanprod(data1_values[i-(n-1):i+1,],0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_Percentile(data1,n,percent):
    #percent=15
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        result[i,:]=np.percentile(data1_values[i-(n-1):i+1,],percent,axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_Percentile_Inv(data1,n):
    data1_values=data1.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        for j in range(data1.shape[1]):
            try:
                result[i,j]=(np.where(np.sort(data1_values[i-(n-1):i+1,j])==data1_values[i,j])[0][0]+1)/n
            except:
                pass
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Markov_Distance(data1,n):
    if n==1:
        result=copy.deepcopy(data1)
    else:
        data1_values=copy.deepcopy(data1.values)
        result=np.full(data1.shape,np.nan)
        start=n-1
        end=data1.shape[0]
        for i in tqdm(range(start,end)):
            temp=data1_values[i-n+1:i+1,]
            temp[np.isnan(temp)]=0
            Sigma=np.cov(temp)
            Value=np.linalg.det(Sigma)
            if (Value==0)|(math.isnan(Value)):
                result[i,:]=0
            else:
                Sigma_Inv=np.linalg.inv(Sigma)
                A=data1_values[i-n+1:i+1,]
                B=np.nanmean(data1_values[i-n+1:i+1,],axis=1)
                AB_DIFF=A-np.tile(B,(data1.shape[1],1)).T
                MAT=AB_DIFF.T.dot(Sigma_Inv).dot(AB_DIFF)
                result[i,:]=np.sqrt(MAT.diagonal())

        where_are_inf = np.isinf(result)

        result[where_are_inf] = np.nan
        result[np.isnan(data1)] = np.nan
        result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

####################横截面####################
def Cs_Single_UpZeroLimit(data1):
    result = data1.where(data1 >= 0 , 0)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Cs_Single_DownZeroLimit(data1):
    result = data1.where(data1 <= 0 , 0)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Cs_ZScore(data1):
    mean_row = np.nanmean(data1,axis = 1)
    std_row = np.nanstd(data1,axis = 1)
    result = data1.subtract(mean_row,axis = 0).divide(std_row,axis =0)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Cs_RankScale(data1):
    result=data1[data1!=0].rank(axis=1)
    result = result.divide(np.nanmax(result,axis = 1),axis = 0).values
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    # result[result == 0] = np.nanmean(result[result != 0],axis = 1)
    for i in range(result.shape[0]):
        temp=result[i,:]
        temp[temp==0]=np.nanmean(temp[temp!=0])
        result[i,:]=temp

    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_DeepScale(data1):
    z_score = data1.subtract( np.nanmean(data1,axis = 1),axis = 0).\
        divide(np.nanstd(data1,axis = 1),axis = 0)
    result = z_score.subtract(np.nanmin(z_score,axis =1 ),axis = 0).\
        divide(np.nanmax(z_score,axis = 1)-np.nanmin(z_score,axis = 1),axis = 0)
    result = result.where(~np.isinf(result), np.nan)

    return result

def Cs_Instant_Increase(data1,ranges):
    result = (np.abs(data1.divide( data1.shift(1))) >= ranges).astype(int)
    # result = np.abs(data1).divide(data1.shift(1))
    result = result.where(~np.isinf(result), np.nan)

    return result

def Cs_Instant_Decrease(data1,ranges):
    result = (np.abs(data1.divide( data1.shift(1))) <= ranges).astype(int)
    result = result.where(~np.isinf(result), np.nan)

    return result

# ===min_periods取1=====
def Cs_Single_RankSD(data1,n):
    result = data1[data1!=0].rank(axis=1)
    result = result.rolling(n,min_periods = 1).std(ddof = 0)
    result = result.where(~np.isinf(result), np.nan)

    return result

def Cs_Single_RankMean(data1,n):
    result = data1[data1!=0].rank(axis=1)
    result = result.rolling(n,min_periods=1).mean()
    result = result.where(~np.isinf(result), np.nan)

    return result

## todo CS 尚未改进
def Cs_Single_Rank(data1):
    result=copy.deepcopy(data1[data1!=0].rank(axis=1).values)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    for i in range(result.shape[0]):
        temp=result[i,:]
        temp[temp==0]=np.nanmean(temp[temp!=0])
        result[i,:]=temp
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Single_Correlation(data1,n):
    # 时序上每一列与所有列均值的相关系数
    if n==1:
        result=copy.deepcopy(data1)
    else:
        data1_values=data1.values
        result=np.full(data1.shape, np.nan)
        result = data1.values.copy()
        start=n-1
        end=data1.shape[0]
        for i in tqdm(range(start,end)):
            M=np.nanmean(data1_values[i-n+1:i+1,],axis=1)
            temp1 = np.tile(M,(data1.shape[1],1)).T
            temp2 = data1_values[i-(n-1):i+1,:]
            result[i,:]=np.nanmean(((temp1 - np.nanmean(temp1,axis = 0))*(temp2-np.nanmean(temp2,axis = 0))),axis = 0)/((np.nanstd(temp1,axis = 0 )*np.nanstd(temp2,axis = 0)))
      
        where_are_inf = np.isinf(result)
      
        result[where_are_inf] = np.nan
        result[np.isnan(data1)] = np.nan
        result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result
        

####################自身运算####################
def Self_Sign(data1):
    result = np.sign(data1)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Self_Negative(data1):
    result=-copy.deepcopy((data1))
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1), np.nan)
    return result

def Self_Inv(data1):
    result=copy.deepcopy(1/data1)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    return result

def Self_Abs(data1):
    result=data1.abs()
    result = result.where(~np.isinf(result), np.nan)
    return result

def Self_Log(data1):
    try:
        result=copy.deepcopy(np.log(data1))

        where_are_inf = np.isinf(result)
    
        #result[where_are_inf] = np.nan
        #result[np.isnan(data1)] = np.nan
    except:
        result=copy.deepcopy(np.log(data1))
    return result

def Self_Sqrt(data1):
    result=copy.deepcopy(np.sqrt(data1))
    result = result.where(~np.isinf(result), np.nan)

    return result

def Self_Sin(data1,n):
    result=copy.deepcopy(np.sin(data1/n/np.pi))
    result = result.where(~np.isinf(result), np.nan)

    return result

def Self_Cos(data1,n):
    result=copy.deepcopy(np.cos(data1/n/np.pi))
    result = result.where(~np.isinf(result), np.nan)

    return result

def Self_Tan(data1,n):
    result=copy.deepcopy(np.tan(data1/n/np.pi))
    result = result.where(~np.isinf(result), np.nan)

    return result

def Self_Power(data1,n):
    result=copy.deepcopy(data1**n)
    result = result.where(~np.isinf(result), np.nan)

    return result

def Self_SignPower(data1,n):
    result=np.sign(data1)*np.power(data1,n)
    result = result.where(~np.isinf(result), np.nan)

    return result

# def test1(data1):
#     result=copy.deepcopy(data1)
#   
#     where_are_inf = np.isinf(result)
#   
#     result[where_are_inf] = np.nan
#     result[np.isnan(data1)] = np.nan
#     return result
#
# def test2(data1):
#     result = data1# copy.deepcopy(data1)
#     result = result.where(~np.isinf(result), np.nan)
#   
#     result = result.where(~np.isinf(result), np.nan)
#     return result

def Self_SignPower_Inv(data1,n):
    result=np.sign(data1)*np.power(data1,1/n)
    result = result.where(~np.isinf(result), np.nan)

    return result

####################Compound Function####################
def Ts_Single_Max_Minus_Min(data1,n):
    rolling_max = data1.rolling(n,min_periods=2).max()
    rolling_min = data1.rolling(n,min_periods=2).min()
    result = rolling_max - rolling_min
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_Max_Plus_Min(data1,n):
    rolling_max = data1.rolling(n,min_periods=2).max()
    rolling_min = data1.rolling(n,min_periods=2).min()
    result = rolling_max + rolling_min
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_Max_Multiply_Min(data1,n):
    rolling_max = data1.rolling(n,min_periods=2).max()
    rolling_min = data1.rolling(n,min_periods=2).min()
    result = rolling_max * rolling_min
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_Max_Divide_Min(data1,n):
    rolling_max = data1.rolling(n,min_periods=2).max()
    rolling_min = data1.rolling(n,min_periods=2).min()
    result = rolling_max / rolling_min
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_Amplitude(data1,n):
    rolling_max = data1.rolling(n,min_periods=2).max()
    rolling_min = data1.rolling(n,min_periods=2).min()
    result = ((rolling_max - rolling_min) / rolling_min).abs()
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Single_Half_Decay(data1,n):
    result=np.full(data1.shape,np.nan)
    data1_values=data1.values
    start=n-1
    end=data1.shape[0]
    wt = [(1/2)**i for i in range(n,0,-1)]
    for i in tqdm(range(start,end)):
        result[i,:]=np.dot(wt ,data1_values[i-n+1:i+1,])
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_Log_Decay(data1,n):
    result=np.full(data1.shape,np.nan)
    data1_values=data1.values
    start=n-1
    end=data1.shape[0]
    wt = np.array( [np.log(i + 2) for i in range(n)] )
    wt = wt / np.sum(wt)
    for i in tqdm(range(start,end)):
        result[i,:]=np.matmul(wt,data1_values[i-n+1:i+1,])
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_Linear_Decay(data1,n):
    result=np.full(data1.shape,np.nan)
    data1_values=data1.ffill().values
    start=n-1
    end=data1.shape[0]
    wt = np.array([range(1,n+1)])
    for i in tqdm(range(start,end)):
        result[i,:]=np.dot(wt,data1_values[i-n+1:i+1,])*2/(n+1)/n
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Regression_SST(data1,n):
    data1_values=data1.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      x1=i-n+1
      y1=i+1
      result[i,:]= np.sum( np.power((data1_values[x1:y1,:]-np.nanmean(data1_values[x1:y1,:], axis = 0)),2), axis = 0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

####################Two####################
# data2的index与columns应当与data1一致，否则强行转换成一致

def Ts_Double_WegihtAverage(data1,data2,n):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    rolling_multiply = (data1 * data2_copy).rolling(n,min_periods=2).sum()
    result = rolling_multiply / data2_copy.rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)

    return result

def Cs_Double_Plus(data1,data2):
    result = data1+data2
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result

    data1_values=data1.values
    data2_values=data2.values
    result=data1_values+data2_values
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Double_Minus(data1,data2):
    result = data1-data2
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result

    data1_values=data1.values
    data2_values=data2.values
    result=data1_values-data2_values
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Double_Multiply(data1,data2):
    result = data1*data2
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result
    
    data1_values=data1.values
    data2_values=data2.values
    result=data1_values*data2_values
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Double_Divide(data1,data2):
    result = data1/data2
    result = result.replace({np.inf: np.nan, -np.inf: np.nan})
    return result

    data1_values=data1.values
    data2_values=data2.values
    result=data1_values/data2_values

    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_UpGap(data1,data2):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = (data1.shift(1) < data2_copy).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Cs_DownGap(data1,data2):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = (data1 < data2_copy.shift(1)).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Ts_Double_Skewness(data1,data2,n):
    data1_values = data1.values
    data1_mean_values=data1.rolling(n,min_periods=2).mean().values
    data1_std_values = data1.rolling(n,min_periods=2).std(ddof = 0).values
    data2_values=data2.values
    data2_mean_values = data2.rolling(n,min_periods=2).mean().values
    data2_std_values = data2.rolling(n,min_periods=2).std(ddof = 0).values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        X=data1_values[i-n+1:i+1,]-data1_mean_values[i,:]
        Y=(data2_values[i-n+1:i+1,]-data2_mean_values[i,:])**2
        XY=(X*Y).mean(axis=0)
        X_SD=data1_std_values[i,:]
        Y_SD=(data2_std_values[i,:])**2
        XY_SD=X_SD*Y_SD
        XY_SD = np.where(XY_SD==0,np.nan,XY_SD)
        result[i,]=XY/(XY_SD)

    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Kurtosis(data1,data2,n):
    data1_values = data1.values
    data1_mean_values=data1.rolling(n,min_periods=2).mean().values
    data1_std_values = data1.rolling(n,min_periods=2).std(ddof = 0).values
    data2_values=data2.values
    data2_mean_values = data2.rolling(n,min_periods=2).mean().values
    data2_std_values = data2.rolling(n,min_periods=2).std(ddof = 0).values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        X=data1_values[i-n+1:i+1,]-data1_mean_values[i,:]
        Y=(data2_values[i-n+1:i+1,]-data2_mean_values[i,:])**3
        XY=(X*Y).mean(axis=0)
        X_SD=data1_std_values[i,:]
        Y_SD=(data2_std_values[i,:])**3
        XY_SD=X_SD*Y_SD
        XY_SD = np.where(XY_SD==0,np.nan,XY_SD)
        result[i,]=XY/(XY_SD)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Euclidean_Distance(data1,data2,n):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = ( (data1 - data2_copy) ** 2 ).rolling(n,min_periods=2).sum() ** 0.5
    result = result.where(~np.isinf(result), np.nan)

    return result

def Ts_Manhattan_Distance(data1,data2,n):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = ( (data1 - data2_copy).abs() ).rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result
# todo 原函数有问题，绝对值取的位置有误
def Ts_Chebyshev_Distance(data1,data2,n):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = ( (data1 - data2_copy).abs() ** n ).rolling(n,min_periods=2).sum() ** (1/n)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Cs_Double_Highest(data1,data2):
    data1_values=data1.values
    data2_values=data2.values
    result = np.nanmax([data1_values,data2_values],axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Double_Lowest(data1,data2):
    data1_values=data1.values
    data2_values=data2.values
    result = np.nanmin([data1_values,data2_values],axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Double_Convergence(data1,data2,diff):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = (data1 - data2_copy).divide(data2_copy).abs()
    result = result.where(result<=diff, 0)
    result = result.where(result <= 0, 1)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Cs_Double_UpCross(data1,data2):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = ((data1.shift(1) < data2_copy.shift(1)) & (data1 > data2_copy)).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Cs_Double_DownCross(data1,data2):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    result = ((data1.shift(1) > data2_copy.shift(1)) & (data1 < data2_copy)).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result


# todo 回归相关尚未优化

def Ts_Double_Correlation(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1 = data1_values[i-(n-1):i+1,:]
        temp2 = data2_values[i-(n-1):i+1,:]
        num = np.nanmean(((temp1 - np.nanmean(temp1,axis = 0))*(temp2-np.nanmean(temp2,axis = 0))),axis = 0)
        den = ((np.nanstd(temp1,axis = 0 )*np.nanstd(temp2,axis = 0)))
        den = np.where(den==0,np.nan,den)
        result[i,:] = num/den
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Spearman(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1=np.argsort(np.argsort(data1_values[i-n+1:i+1,:],0),0)
        temp2=np.argsort(np.argsort(data2_values[i-n+1:i+1,:],0),0)
        temp=np.nansum((temp1-temp2)**2,0)
        result[i,:]=1-6*temp/(n)/(n**2-1)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Regression_Beta0(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[0]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Regression_Beta1(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[1]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Regression_FV(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[0]+beta[1]*data2_values[i,j]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Regression_Res(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[0]+beta[1]*data2_values[i,j]-data1_values[i,j]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Regression_SSE(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=pow((beta[0]+beta[1]*data2_values[x1:y1,j]-data1_values[x1:y1,j]),2).sum()
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Regression_SSR(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=pow((beta[0]+beta[1]*data2_values[x1:y1,j]-np.nanmean(data1_values[x1:y1,j])),2).sum()
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Double_Regression_Rsuqre(data1,data2,n):
    data1_values=data1.values
    data2_values=data2.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          SSR=pow((beta[0]+beta[1]*data2_values[x1:y1,j]-np.nanmean(data1_values[x1:y1,j])),2).sum()
          SST=pow((data1_values[x1:y1,j]-np.nanmean(data1_values[x1:y1,j])),2).sum()
          result[i,j]=SSR/SST
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

####################Three####################
# data2,data3的index与columns应当与data1一致，否则强行转换成一致
def Cs_Triple_Plus(data1,data2,data3):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=data1_values+data2_values+data3_values
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Triple_Multiply(data1,data2,data3):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=data1_values*data2_values*data3_values
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Triple_Highest(data1,data2,data3):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result = np.nanmax([data1_values,data2_values,data3_values],axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Triple_Lowest(data1,data2,data3):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result = np.nanmin([data1_values,data2_values,data3_values],axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Triple_UpCross(data1,data2,data3):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    data3_copy = data3.copy()
    data3_copy.columns = data1.columns
    result = ((data1.shift(1) < data2_copy.shift(1)) & (data1 > data2_copy) & \
              (data1.shift(1) < data3_copy.shift(1)) & (data1 > data3_copy) ).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Cs_Triple_DownCross(data1,data2,data3):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    data3_copy = data3.copy()
    data3_copy.columns = data1.columns
    result = ((data1.shift(1) > data2_copy.shift(1)) & (data1 < data2_copy) & \
              (data1.shift(1) > data3_copy.shift(1)) & (data1 < data3_copy) ).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

# todo 回归相关尚未优化
def Ts_Triple_Regression_Beta0(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[0]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Regression_Beta1(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[1]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Regression_Beta2(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[2]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Regression_FV(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[0]+beta[1]*data2_values[i,j]+beta[2]*data3_values[i,j]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Regression_Res(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=beta[0]+beta[1]*data2_values[i,j]+beta[2]*data3_values[i,j]-data1_values[i,j]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Regression_SSE(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=pow((beta[0]+beta[1]*data2_values[x1:y1,j]+beta[2]*data3_values[x1:y1,j]-data1_values[x1:y1,j]),2).sum()
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Regression_SSR(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          result[i,j]=pow((beta[0]+beta[1]*data2_values[x1:y1,j]+beta[2]*data3_values[x1:y1,j]-np.nanmean(data1_values[x1:y1,j])),2).sum()
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Regression_Rsuqre(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
      for j in range(data1.shape[1]):
        x1=i-n+1
        y1=i+1
        Y=data1_values[x1:y1,j]
        X=np.vstack((np.ones((1,n)),data2_values[x1:y1,j],data3_values[x1:y1,j])).T
        Mat=np.matmul(X.T,X)
        Value=np.linalg.det(Mat)
        if (Value==0)|(math.isnan(Value)):
          result[i,j]=0
        else:
          beta=np.linalg.inv(Mat).dot(X.T).dot(Y)
          SSR=pow((beta[0]+beta[1]*data2_values[x1:y1,j]+beta[2]*data3_values[x1:y1,j]-np.nanmean(data1_values[x1:y1,j])),2).sum()
          SST=pow((data1_values[x1:y1,j]-np.nanmean(data1_values[x1:y1,j])),2).sum()
          result[i,j]=SSR/SST
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Triple_Correlation(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result=np.full(data1.shape,np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp1 = data1_values[i-(n-1):i+1,:]
        temp2 = data2_values[i-(n-1):i+1,:]
        temp3 = data3_values[i-(n-1):i+1,:]
        result[i,:] = np.nanmean(((temp1 - np.nanmean(temp1,axis = 0))*(temp2-np.nanmean(temp2,axis = 0))*(temp3-np.nanmean(temp3,axis = 0))),axis = 0)/((np.nanstd(temp1,axis = 0 )*np.nanstd(temp2,axis = 0))*np.nanstd(temp3,axis = 0))
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

####################其他####################
def Ts_Shape_Doji(data1,data2,data3,data4,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    data4_values=data4.values
    result_value = ((np.abs((data2_values - data1_values) / data1_values) <= 0.02) & (
                np.abs((data2_values - data3_values) / data2_values) >= 0.02) & (
                 np.abs((data2_values - data4_values) / data2_values) >= 0.02)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Ts_Shape_UpLongLineRed(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result_value = ((data2_values>data1_values)&
                    ((data3_values-data2_values)/data2_values>=0.04)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Ts_Shape_DownLongLineRed(data1,data2,data4,n):
    data1_values=data1.values
    data2_values=data2.values
    data4_values=data4.values
    result_value = ((data2_values>data1_values)\
                    &((data1_values-data4_values)/data1_values>=0.04)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Ts_Shape_UpLongLineGreen(data1,data2,data3,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    result_value = ((data2_values<data1_values)\
                    &((data3_values-data1_values)/data1_values>=0.04)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_Shape_DownLongLineGreen(data1,data2,data4,n):
    data1_values=data1.values
    data2_values=data2.values
    data4_values=data4.values
    result_value = ((data2_values<data1_values) \
                    &((data2_values-data4_values)/data1_values>=0.04)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_Shape_PlainDownLongLineRed(data1,data2,data3,data4,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    data4_values=data4.values
    result_value = ((data2_values>data1_values)\
                    &((data1_values-data4_values)/data1_values>=0.04)\
                    &(np.abs((data2_values-data3_values)/data2_values)<=0.01)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isinf(result), np.nan)
    return result

def Ts_Shape_PlainUpLongLineRed(data1,data2,data3,data4,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    data4_values=data4.values
    result_value = ((data2_values>data1_values)\
                    &((data3_values-data2_values)/data2_values>=0.04)\
                    &(np.abs((data1_values-data4_values)/data1_values)<=0.01)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_Shape_PlainUpLongLineGreen(data1,data2,data3,data4,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    data4_values=data4.values
    result_value = ((data2_values<data1_values)\
                    &((data3_values-data1_values)/data1_values>=0.04)\
                    &(np.abs((data2_values-data4_values)/data2_values)<=0.01)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_Shape_PlainDownLongLineGreen(data1,data2,data3,data4,n):
    data1_values=data1.values
    data2_values=data2.values
    data3_values=data3.values
    data4_values=data4.values
    result_value = ((data2_values<data1_values)\
                    &((data2_values-data4_values)/data1_values>=0.04)\
                    &(np.abs((data1_values-data3_values)/data1_values)<=0.01)).astype(int)
    result = pd.DataFrame(result_value, index=data1.index, columns=data1.columns).\
        rolling(n,min_periods=2).sum()
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Cs_Quadruple_UpCross(data1,data2,data3,data4):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    data3_copy = data3.copy()
    data3_copy.columns = data1.columns
    data4_copy = data4.copy()
    data4_copy.columns = data1.columns
    result = ((data1.shift(1) < data2_copy.shift(1)) & (data1 > data2_copy) & \
              (data1.shift(1) < data3_copy.shift(1)) & (data1 > data3_copy) & \
              (data1.shift(1) < data4_copy.shift(1)) & (data1 > data4_copy) ).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Cs_Quadruple_DownCross(data1,data2,data3,data4):
    data2_copy = data2.copy()
    data2_copy.columns = data1.columns
    data3_copy = data3.copy()
    data3_copy.columns = data1.columns
    data4_copy = data4.copy()
    data4_copy.columns = data1.columns
    result = ((data1.shift(1) > data2_copy.shift(1)) & (data1 < data2_copy) & \
              (data1.shift(1) > data3_copy.shift(1)) & (data1 < data3_copy) & \
              (data1.shift(1) > data4_copy.shift(1)) & (data1 < data4_copy) ).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Dummy_Transform(data1):
    result=copy.deepcopy(data1)
    result[result>0]=1
    result[result<0]=0
    #result+=0.01
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    return result

def Ts_Single_RSV(data1,data2,data3,n):
    min_n = data3.rolling(n,min_periods=2).min().values
    max_n = data2.rolling(n,min_periods=2).max().values
    data1_values=data1.values
    result = (data1_values-min_n) /(max_n-min_n)*100
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_K(data1,data2,data3,n):
    K_initial=50
    data1_values=data1.values
    min_n = data3.rolling(n,min_periods=2).min().values
    max_n = data2.rolling(n,min_periods=2).max().values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        K_initial=(2/3)*K_initial+(1/3)*(data1_values[i,:]-min_n[i,:])/(max_n[i,:]-min_n[i,:])*100
        result[i,:]=K_initial
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_D(data1,data2,data3,n):
    K_initial=50
    D_initial=50
    data1_values=data1.values
    min_n = data3.rolling(n,min_periods=2).min().values
    max_n = data2.rolling(n,min_periods=2).max().values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        K_initial=(2/3)*K_initial+(1/3)*(data1_values[i,:]-min_n[i,:])/(max_n[i,:]-min_n[i,:])*100
        D_initial=(2/3)*D_initial+(1/3)*K_initial
        result[i,:]=D_initial
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_J(data1,data2,data3,n):
    K_initial=50
    D_initial=50
    data1_values=data1.values
    min_n = data3.rolling(n,min_periods=2).min().values
    max_n = data2.rolling(n,min_periods=2).max().values
    result=np.full(data1.shape, np.nan)
    start=n-1
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        K_initial=(2/3)*K_initial+(1/3)*(data1_values[i,:]-min_n[i,:])/(max_n[i,:]-min_n[i,:])*100
        D_initial=(2/3)*D_initial+(1/3)*K_initial
        result[i,:]=3*K_initial-2*D_initial
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_UpTrend(data1,n):
    result = (data1.diff() > 0).astype(int)
    result = (result.rolling(n,min_periods=2).sum() == n ).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_DownTrend(data1,n):
    result = (data1.diff() < 0).astype(int)
    result = (result.rolling(n,min_periods=2).sum() == n ).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_Single_HPM(data1,n,limit=0.01):
    maxs = data1.rolling(n,min_periods=2).max()
    mins = data1.rolling(n,min_periods=2).min()
    result = ((maxs-mins)/mins<=limit).astype(int)
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_Single_RoughUpTrend(data1, n, rough_limit=0.75):
    result = np.zeros(data1.shape)
    data1_values = data1.values
    start = n - 1
    end = data1.shape[0]
    parameters = n - 1
    for i in tqdm(range(start, end)):
        temp = data1_values[i - n + 1:i + 1, :]
        total_trend = ((temp[-1] - temp[1]) > 0).astype(int)
        # 向量化运算
        count = np.nansum(np.diff(temp, axis=0) > 0, axis=0)
        #
        rough_trend = (count / parameters >= rough_limit).astype(int)
        result[i, :] += total_trend * rough_trend
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = 0
    result = pd.DataFrame(result, index=data1.index, columns=data1.columns)
    return result

def Ts_Single_RoughDownTrend(data1,n,rough_limit=0.75):
    result=np.zeros(data1.shape)
    data1_values=data1.values
    start=n-1
    end=data1.shape[0]
    parameters=n-1
    for i in tqdm(range(start,end)):
        temp = data1_values[i - n + 1:i + 1, :]
        total_trend = ((temp[-1] - temp[1]) < 0).astype(int)
        #
        count = np.nansum(np.diff(temp, axis=0) < 0, axis=0)
        #
        rough_trend=(count/parameters>=rough_limit).astype(int)
        result[i,:]+=total_trend*rough_trend
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_LinearStretch(data1,r1=0,r2=1):
    length=data1.shape[0]
    ratio=np.tile(np.linspace(r1,r2,length),(data1.shape[1],1)).T
    result=data1*ratio
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

def Ts_Single_NonlinearStretch(data1,r1=0,r2=1,times=0.5):
    length=data1.shape[0]
    ratio=np.tile(np.linspace(r1,r2,length)**times,(data1.shape[1],1)).T
    result=data1*ratio
    result = result.where(~np.isinf(result), np.nan)
    result = result.where(~np.isnan(data1),np.nan)
    return result

# todo 尚未优化
def Ts_Single_SMA(data1,n,M=3):
    ratio1=min(M/n,1)
    ratio2=1-ratio1
    result=np.zeros(data1.shape)
    data1_values=data1.values
    start=n
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        temp=data1_values[i-n+1:i+1,:]
        temp1=temp[0,:]
        for j in range(1,n):
            temp1=ratio2*temp1+ratio1*temp[j]
        result[i,:]=temp1
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Ts_Single_TR(high,low,close):
    result=np.zeros(high.shape)
    data1_values=(high-low).values
    data2_values=abs(low-close.shift(1)).values
    data3_values=abs(high-close.shift(1)).values
    start=0
    end=high.shape[0]
    for i in tqdm(range(start,end)):
        result[i,:]=np.nanmax([data1_values[i,:],data2_values[i,:],data3_values[i,:]],axis=0)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(high)]=0
    result=pd.DataFrame(result,index=high.index,columns=high.columns)
    return result

def Cs_Single_Regression_Res(data1, data2):
    data1_values = data1.values
    if len(data2.shape) == 1:
        data2_values = np.tile(data2, (data1.shape[0], 1))
    else:
        data2_values = data2.values
    result = np.full(data1.shape, np.nan)
    start = 0
    end = data1.shape[0]
    for i in tqdm(range(start, end)):
        x = data2_values[i]
        y = data1_values[i]

        beta = (x - np.nanmean(x)) * (y - np.nanmean(y)) / np.nanvar(x) / len(x)
        beta0 = np.nanmean(y) - beta * np.nanmean(x)
        res = beta * x + beta0 - y

        result[i] = res
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = 0
    result = pd.DataFrame(result, index=data1.index, columns=data1.columns)
    return result

def Cs_Muti_Regression_Res(data1, data2):
    data1 = data1.loc[~((data1 == 0) | (np.isnan(data1))).all(axis=1)]
    both_stock = data1.columns & data2.index
    data1_values = data1.loc[:, both_stock].values
    data2_values = data2.loc[both_stock].values
    result = np.full(data1_values.shape, np.nan)
    start = 0
    end = data1.shape[0]
    for i in tqdm(range(start, end)):
        Y = data1_values[i]
        X = data2_values
        delete = (np.isnan(Y))
        Y_ = Y[~delete]
        X_ = X[~delete, :]
        slr = LinearRegression()
        slr.fit(X_, Y_)
        result[i] = slr.predict(X) - Y
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result = pd.DataFrame(result, index=data1.index, columns=both_stock)
    return result

def Cs_Double_Regression_Res(data1, data2):
    data1_values = data1.values
    data2_values = data2.values
    result = np.full(data1_values.shape, np.nan)
    start = 0
    end = data1.shape[0]
    for i in tqdm(range(start, end)):
        try:
            Y = data1_values[i]
            X = data2_values[i]
            delete = (np.isnan(Y) | (Y == 0))
            Y_ = Y[~delete]
            X_ = X[~delete].reshape(-1, 1)
            slr = LinearRegression()
            slr.fit(X_, Y_)
            result[i] = slr.predict(X.reshape(-1, 1)) - Y
        except:
            pass
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result = pd.DataFrame(result, index=data1.index, columns=data1.columns)
    return result

def Cs_Fill_Stock(data1,data2):
    both_index=data1.index&data2.index
    data1=data1.loc[both_index]
    data2=data2.loc[both_index]
    result=pd.DataFrame(np.tile(data1,(data2.shape[1],1)).T,index=data2.index,columns=data2.columns)
    return result

def Ts_Fill_Time(data1,data2):
    both_index=data1.index&data2.columns
    data1=data1.loc[both_index]
    data2=data2.loc[:,both_index]
    result=pd.DataFrame(np.tile(data1,(data2.shape[0],1)),index=data2.index,columns=data2.columns)
    return result

def Cs_Extremum_Mad(data1,n=20,axis=1):
    data=copy.deepcopy(data1.values)
    #####考虑一维情况
    if len(data.shape)==1:
        temp=data
        median_temp=np.nanmedian(temp)
        new_median=np.nanmedian(abs(temp-median_temp))
        data=np.clip(temp,median_temp-n*new_median,new_median+n*new_median)
        data=pd.Series(data,index=data1.index)
    else:
        if axis==1:
            for i in range(data.shape[0]):
                temp=data[i]
                median_temp=np.nanmedian(temp)
                new_median=np.nanmedian(abs(temp-median_temp))
                data[i]=np.clip(temp,median_temp-n*new_median,new_median+n*new_median)
        elif axis==0:
            for i in range(data.shape[1]):
                temp=data[:,i]
                median_temp=np.nanmedian(temp)
                new_median=np.nanmedian(abs(temp-median_temp))
                data[:,i]=np.clip(temp,median_temp-n*new_median,new_median+n*new_median)
        data=pd.DataFrame(data,data1.index,data1.columns)
    return data
# 存疑
#data1=crossStockData['close1']
#n=10
#a=Ts_Single_Cummax(data1,n)
#data1.rolling(n,min_periods=2).sum()
def Ts_Single_Cummax(data1,n=np.inf):
    if n==np.inf:
        result=copy.deepcopy(data1.cummax(0))
    else:
        result=np.full(data1.shape,np.nan)
        start=n-1
        end=data1.shape[0]
        for i in tqdm(range(start,end)):
            result[i,:]=data1.iloc[i-n+1:i+1,:].cummax(0).iloc[-1]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result
def Ts_Single_Cummin(data1,n=np.inf):
    if n==np.inf:
        result=copy.deepcopy(data1.cummin(0))
    else:
        result=np.full(data1.shape,np.nan)
        start=n-1
        end=data1.shape[0]
        for i in tqdm(range(start,end)):
            result[i,:]=data1.iloc[i-n+1:i+1,:].cummin(0).iloc[-1]
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result[np.isnan(data1)] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Score_Adjusted(data1,n=10,Decreasing=True):
    n=min(n,len(data1))
    if len(data1)==0:
        temp=copy.deepcopy(data1)
    else:
        data=copy.deepcopy(data1)
        data[data.isnull()]=0
        data=data.fillna(0)
        data=data.replace([np.inf, -np.inf],0)
        data[data==0]=np.nanmedian(data[data!=0])
        standard=np.linspace(0.1,1,n)
        standard_cutoff=np.linspace(0,1,n+1)
        cut_off_postion=np.round(len(data)*standard_cutoff)
        temp= data.sort_values(ascending=Decreasing)
        for i in range(n):
            temp.iloc[int(cut_off_postion[i]):int(cut_off_postion[i+1])]=standard[i]
        temp=temp.sort_index()
    return temp

def Score_Adjusted_Specific(data1,n=10,Decreasing=True):
    data_o=pd.Series(0,index=data1.index)
    data=data1[(data1!=0)&(~np.isnan(data1))&(~np.isinf(data1))]
    n=min(n,len(data))
    standard=np.linspace(0.1,1,n)
    standard_cutoff=np.linspace(0,1,n+1)
    cut_off_postion=np.round(len(data)*standard_cutoff)
    temp= data.sort_values(ascending=Decreasing)
    for i in range(n):
        temp.iloc[int(cut_off_postion[i]):int(cut_off_postion[i+1])]=standard[i]
    temp=(temp-np.nanmean(temp))/np.nanstd(temp)
    data_o.loc[temp.index&data_o.index]=temp.loc[temp.index&data_o.index]
    return data_o

def Cs_Single_Stratify(data1,layers=10):
    result=np.full(data1.shape,np.nan)
    data1_values=copy.deepcopy(data1)
    start=0
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        result[i]=Score_Adjusted(data1_values.iloc[i],n=layers)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Single_Stratify_Specific(data1,layers=10):
    result=np.full(data1.shape,np.nan)
    data1_values=copy.deepcopy(data1)
    start=0
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        result[i]=Score_Adjusted_Specific(data1_values.iloc[i],n=layers)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Self_Stratify(data1):
    layers=10
    result=np.full(data1.shape,np.nan)
    data1_values=copy.deepcopy(data1)
    start=0
    end=data1.shape[0]
    for i in tqdm(range(start,end)):
        result[i]=Score_Adjusted(data1_values.iloc[i],n=layers)
    where_are_inf = np.isinf(result)
    result[where_are_inf] = np.nan
    result=pd.DataFrame(result,index=data1.index,columns=data1.columns)
    return result

def Cs_Industry_Stratify(data1,industry_matrix,layer=10):
    result=copy.deepcopy(data1)
    data1_values=copy.deepcopy(data1)
    start=0
    end=data1.shape[0]
    for i in tqdm(data1.index):
        stock_temp=data1_values.loc[i]
        for j in industry_matrix.columns:
            temp=industry_matrix[j]
            in_index=(temp[temp==1].index)&(data1.columns)
            result.loc[i,in_index]=Score_Adjusted(stock_temp.loc[in_index],n=layer)
    return result

def Cs_Single_Industry_Stratify(data1):
    ##########industry
    layer=10
    result=copy.deepcopy(data1)
    data1_values=copy.deepcopy(data1)
    start=0
    end=data1.shape[0]
    # for i in tqdm(data1.index):
    #     stock_temp=data1_values.loc[i]
    #     for j in industry_matrix.columns:
    #         temp=industry_matrix[j]
    #         in_index=(temp[temp==1].index)&(data1.columns)
    #         result.loc[i,in_index]=Score_Adjusted(stock_temp.loc[in_index],n=layer)
    return result

